import re

from m.ci.versioning import VersionInputs, build_m_tag, build_py_tag
from m.core.maybe import maybe
from m.log import EnvVars
from pydantic import BaseModel

from ..core import Res, issue
from ..core.fp import Bad, Good
from ..github.ci import (
    Commit,
    CommitInfo,
    PullRequest,
    Release,
    get_ci_run_info,
)
from .config import Config


def get_release_prefix(config: Config) -> str | None:
    """Find out the release prefix.

    Args:
        config:
            The m configuration. Its workflow is used to determine the prefix.

    Returns:
        The release prefix or None if not using a supported workflow.
    """
    if config.uses_git_flow():
        return config.git_flow.release_prefix
    if config.uses_m_flow():
        return config.m_flow.release_prefix
    return None


def get_hotfix_prefix(config: Config) -> str | None:
    """Find out the hotfix prefix.

    Args:
        config:
            The m configuration. Its workflow is used to determine the prefix.

    Returns:
        The hofix prefix or None if not using a supported workflow.
    """
    if config.uses_git_flow():
        return config.git_flow.hotfix_prefix
    if config.uses_m_flow():
        return config.m_flow.hotfix_prefix
    return None


def _version_inputs(
    git_env: 'GitEnv',
    config: Config,
    run_id: str,
) -> VersionInputs | None:
    is_release = git_env.is_release(config)
    is_release_pr = git_env.is_release_pr(config)
    is_hotfix_pr = git_env.is_hotfix_pr(config)
    is_dev_branch = git_env.target_branch == config.git_flow.develop_branch
    if config.uses_git_flow() and is_dev_branch:
        if is_release or is_release_pr or is_hotfix_pr:
            return None
    pr_number = maybe(lambda: git_env.pull_request.pr_number)  # type: ignore[union-attr]
    return VersionInputs(
        version=config.version,
        version_prefix=_build_tag_prefix(config),
        run_id=run_id,
        sha=git_env.sha,
        pr_number=pr_number,
        branch=git_env.target_branch,
        is_release=is_release,
        is_release_pr=is_release_pr,
        is_hotfix_pr=is_hotfix_pr,
    )


class GitEnv(BaseModel):
    """Object to store the git configuration."""

    sha: str
    branch: str
    target_branch: str
    commit: Commit | None = None
    pull_request: PullRequest | None = None
    release: Release | None = None

    def get_pr_branch(self) -> str:
        """Get the pull request branch or empty string.

        Returns:
            The name of the pull request branch or an empty string when not
            dealing with a pull request.
        """
        return self.pull_request.pr_branch if self.pull_request else ''

    def get_pr_number(self) -> int:
        """Get the pull request number or 0 if not a pull request.

        Returns:
            The pull request number or 0 if not a pull request.
        """
        return self.pull_request.pr_number if self.pull_request else 0

    def is_release(self, config: Config) -> bool:
        """Determine if the current commit should create a release.

        Args:
            config: The `m` configuration.

        Returns:
            True if we are dealing with a release.
        """
        if not self.commit:
            return False
        release_prefix = get_release_prefix(config)
        hotfix_prefix = get_hotfix_prefix(config)
        if config.uses_m_flow() and self.branch != config.m_flow.master_branch:
            return False
        if config.uses_git_flow():
            if self.branch != config.git_flow.master_branch:
                return False
        return (
            self.commit.is_release(release_prefix)
            or self.commit.is_release(hotfix_prefix)
        )

    def is_release_pr(self, config: Config) -> bool:
        """Determine if the the current pr is a release pr.

        Args:
            config: The `m` configuration.

        Returns:
            True if we are dealing with a release pr.
        """
        if not self.pull_request:
            return False
        release_prefix = get_release_prefix(config)
        return self.pull_request.is_release_pr(release_prefix)

    def is_hotfix_pr(self, config: Config) -> bool:
        """Determine if the the current pr is a hotfix pr.

        It is a release pr as far as the pull request should see it but
        from the context of the git environment we need to label it as a
        hotfix pr.

        Args:
            config: The m configuration object.

        Returns:
            True if the we are dealing with hotfix pr.
        """
        if not self.pull_request:
            return False
        hotfix_prefix = get_hotfix_prefix(config)
        return self.pull_request.is_release_pr(hotfix_prefix)

    def get_build_tag(self, config: Config, run_id: str) -> Res[str]:
        """Create a build tag for the current commit.

        It is tempting to use the config_version when creating a build tag for
        pull requests or branches. This will only be annoying when testing.

        Consider the following scenario. An application is being tested with
        `1.0.1-pr99.b123`. When using docker you may want to refer to the
        latest pr build by using `1.0.1-pr99`. Now lets say that a release
        happened and now the config_version is at `1.1.0`. The application
        build will not get the latest changes because the new changes are in
        `1.1.0-pr99`.

        There are two solutions, either always state the version that is
        being used or make a tag to depend only on the pull request number.
        This is the reason why for prs (constantly changing) we avoid
        using the version in the configuration.

        For release prs we use `rc` followed by the pull request. In this case
        it is safe to use config_version given that there should only be
        one release at a time. We treat hotfixes similar to releases.

        Git flow will generate a special build tag: SKIP. This will happen when
        we try to merge a release or hotfix branch to the develop branch.

        Args:
            config: The `m` configuration.
            run_id: A unique identifier for the run/job.

        Returns:
            A unique identifier for the build. This tag is compatible with
            both `npm` and `docker`.
        """
        ver_input = _version_inputs(self, config, run_id)
        if ver_input:
            return Good(build_m_tag(ver_input, config))
        return Good('SKIP')

    def get_py_tag(self, config: Config, run_id: str) -> Res[str]:
        """Create a python tag for the current commit.

        Args:
            config: The `m` configuration.
            run_id: A unique identifier for the run/job.

        Returns:
            A unique identifier for the build. This tag is compatible with
            python.
        """
        ver_input = _version_inputs(self, config, run_id)
        if ver_input:
            return Good(build_py_tag(ver_input, config))
        return Good('SKIP')


def _build_tag_prefix(config: Config) -> str:
    if config.uses_free_flow():
        return ''
    if config.build_tag_with_version:
        ver = config.version.split('-')[0]
        parts = [int(x) for x in ver.split('.')]
        minor_ver = parts[1] + 1
        return f'{parts[0]}.{minor_ver}.0'
    return '0.0.0'


def get_pr_number(branch: str) -> int | None:
    """Retrieve the pull request number from the branch name.

    Args:
        branch: The branch name from where the pr number is extracted.py

    Returns:
        The pr number if the branch is a pull request otherwise `None`.
    """
    if 'pull/' in branch:
        parts = branch.split('/')
        return int(parts[parts.index('pull') + 1])
    return None


def _remove_strings(str_content: str, words: list[str]) -> str:
    return re.sub('|'.join(words), '', str_content)


def get_git_env(config: Config, env_vars: EnvVars) -> Res[GitEnv]:
    """Obtain the git environment by asking Github's API.

    Args:
        config: The m configuration object.
        env_vars: The environment variables.

    Returns:
        The git environment object or an issue.
    """
    branch = _remove_strings(env_vars.git_branch, ['refs/heads/', 'heads/'])
    sha = env_vars.git_sha
    git_env = GitEnv(sha=sha, branch=branch, target_branch=branch)

    # quick exit for local environment
    if not env_vars.ci_env:
        return Good(git_env)

    pr_number = get_pr_number(branch)
    git_env_box = get_ci_run_info(
        token=env_vars.github_token,
        commit_info=CommitInfo(
            owner=config.owner,
            repo=config.repo,
            sha=env_vars.git_sha,
        ),
        pr_number=pr_number,
        file_count=0,
        include_release=True,
    )
    if isinstance(git_env_box, Bad):
        return issue('git_env failure', cause=git_env_box.value)

    res = git_env_box.value
    pr = res.pull_request
    git_env.sha = res.commit.sha
    git_env.target_branch = pr.target_branch if pr else branch
    git_env.commit = res.commit
    git_env.pull_request = res.pull_request
    git_env.release = res.release
    return Good(git_env)



import os
import shutil
from pathlib import Path

from m.core import Bad, Good, Res, issue
from m.core import subprocess as sub
from m.log import Logger
from pydantic import BaseModel

logger = Logger('m.devcontainer.pnpm')

# list of commands that need to be executed in the mounted volume
PNPM_MOUNTED_COMMANDS = (
    'add',
    'i',
    'install',
    'ln',
    'link',
    'prune',
    'rm',
    'remove',
    'unlink',
    'up',
    'update',
)
SUGGESTION = 'suggestion'


def create_symlink(link: Path, source: Path) -> None:
    """Create a symlink that is linked to a source.

    This is a destructive operation - if the link exists it will be removed.

    Args:
        link: The path to the link.
        source: The path to the source.
    """
    if link.exists() or link.is_symlink():
        link.unlink()
    link.symlink_to(source)


class PnpmSetupSummary(BaseModel):
    """Summary of the pnpm setup operation."""

    node_modules: str
    package: str
    npmrc: str
    pnpm_lock: str | None


def _setup_node_modules(work_dir: str, pnpm_dir: str) -> Res[str]:
    work_node_modules = Path(work_dir) / 'node_modules'
    pnpm_node_modules = Path(pnpm_dir) / 'node_modules'

    if work_node_modules.exists() and not work_node_modules.is_symlink():
        # if we remove the directory from inside the container it will take
        # a while, better let the developer do it from the host.
        return issue('non_devcontainer_setup', context={
            'work_node_modules': str(work_node_modules),
            'pnpm_node_modules': str(pnpm_node_modules),
            SUGGESTION: [
                'stop devcontainer',
                'remove all node_modules found in the project directory',
                'reopen project in devcontainer',
            ],
        })

    pnpm_node_modules.mkdir(exist_ok=True)
    create_symlink(work_node_modules, pnpm_node_modules)
    return Good(f'{work_node_modules} -> {pnpm_node_modules}')


def _setup_package(work_dir: str, pnpm_dir: str) -> Res[str]:
    work_package = Path(work_dir) / 'package.json'
    pnpm_package = Path(pnpm_dir) / 'package.json'
    if not work_package.exists():
        return issue('missing_package_json', context={
            'work_dir': str(work_dir),
            SUGGESTION: [
                'directory does not have package.json',
                'you may create one with pnpm init',
            ],
        })
    # ensure the pnpm directory exists otherwise symlink will fail
    Path(pnpm_dir).mkdir(parents=True, exist_ok=True)
    create_symlink(pnpm_package, work_package)
    return Good(f'{pnpm_package} -> {work_package}')


def _setup_npmrc(work_dir: str, pnpm_dir: str) -> Res[str]:
    npmrc_pnpm = Path(pnpm_dir) / '.npmrc'
    npmrc_work = Path(work_dir) / '.npmrc'
    npmrc_home = Path(os.environ.get('HOME', 'undefined')) / '.npmrc'
    npmrc_path = npmrc_work if npmrc_work.exists() else npmrc_home
    if not npmrc_path.exists():
        return issue('missing_npmrc', context={
            'npmrc_work': f'MISSING {npmrc_work}',
            'npmrc_home': f'MISSING {npmrc_home}',
        })
    create_symlink(npmrc_pnpm, npmrc_path)
    return Good(f'{npmrc_pnpm} -> {npmrc_path}')


def pnpm_setup(work_dir: str, pnpm_dir: str) -> Res[None]:
    """Create symbolic links to a mounted volume.

    This is done so that pnpm may take advantage of the cache associated with
    a single pnpm store. When we execute `pnpm install` in the devcontainer
    we want to make sure that other containers may be able to share the pnpm
    cache. To do this we need all containers to use the same pnpm store.

    Args:
        work_dir: The directory where the project is mounted.
        pnpm_dir: The directory where the project will execute pnpm commands.

    Returns:
        A `PnpmSetupSummary` instance.
    """
    package_res = _setup_package(work_dir, pnpm_dir)
    if isinstance(package_res, Bad):
        return Bad(package_res.value)
    package_summary = package_res.value

    node_modules_res = _setup_node_modules(work_dir, pnpm_dir)
    if isinstance(node_modules_res, Bad):
        return Bad(node_modules_res.value)
    node_modules_summary = node_modules_res.value

    npmrc_res = _setup_npmrc(work_dir, pnpm_dir)
    if isinstance(npmrc_res, Bad):
        return Bad(npmrc_res.value)
    npmrc_summary = npmrc_res.value

    # perform a few checks with the lock file
    work_lock = Path(work_dir) / 'pnpm-lock.yaml'
    pnpm_lock_summary = None
    if not work_lock.exists():
        pnpm_lock_summary = f'MISSING {work_lock}'
        logger.warning('pnpm_lock_missing', context={
            'work_lock': pnpm_lock_summary,
            SUGGESTION: 'run `pnpm install` to generate the lock file',
        })
    if work_lock.is_symlink():
        pnpm_lock_summary = f'unlinked symlink {work_lock}'
        work_lock.unlink()

    summary = PnpmSetupSummary(
        node_modules=node_modules_summary,
        package=package_summary,
        npmrc=npmrc_summary,
        pnpm_lock=pnpm_lock_summary,
    )
    logger.debug('pnpm_setup_summary', context=summary.model_dump())
    return Good(None)


def _pnpm(work_dir: str, pnpm_dir: str, pnpm_args: list[str]) -> Res[None]:
    """Execute the pnpm command by first switching to the given pnpm directory.

    Args:
        work_dir: The directory where the project is mounted.
        pnpm_dir: The directory where the project will execute pnpm commands.
        pnpm_args: The arguments to pass to pnpm.

    Returns:
        None if successful, otherwise an error message.
    """
    error = None
    if not Path(f'{work_dir}/node_modules').is_symlink():
        error = f'{work_dir}/node_modules is not a symlink'
    elif not Path(f'{pnpm_dir}/.npmrc').is_symlink():
        error = f'{pnpm_dir}/.npmrc is not a symlink'
    if error:
        return issue('invalid_pnpm_setup', context={'error': error})

    os.chdir(pnpm_dir)
    store_path_res = sub.eval_cmd('pnpm store path')
    if isinstance(store_path_res, Bad):
        return Bad(store_path_res.value)
    logger.info('executing_pnpm_in_mounted_volume', context={
        'store_path': store_path_res.value,
        'pnpm_dir': pnpm_dir,
    })

    pnpm_res = sub.exec_pnpm(pnpm_args)
    if isinstance(pnpm_res, Bad):
        return pnpm_res

    work_lock = Path(work_dir) / 'pnpm-lock.yaml'
    pnpm_lock = Path(pnpm_dir) / 'pnpm-lock.yaml'
    if not work_lock.exists():
        shutil.move(str(pnpm_lock), str(work_lock))
    create_symlink(pnpm_lock, work_lock)
    return Good(None)


def _get_workspaces(workdir: str) -> Res[tuple[str, str]]:
    workspace = os.environ.get('MDC_WORKSPACE')
    pnpm_workspace = os.environ.get('MDC_PNPM_WORKSPACE')
    if not workspace or not pnpm_workspace:
        return issue('missing_env_var', context={
            'MDC_WORKSPACE': workspace,
            'MDC_PNPM_WORKSPACE': pnpm_workspace,
            SUGGESTION: 'are you running this command from a devcontainer?',
        })
    if not workdir.startswith(workspace):
        return issue('invalid_devcontainer_pnpm_use', context={
            'workdir': workdir,
            'workspace': workspace,
            'suggestions': [
                'pnpm alias should only be run in the `workspace`',
                'the original `pnpm` may be used by running `command pnpm`',
            ],
        })
    return Good((workspace, pnpm_workspace))


def run_pnpm(pnpm_args: list[str], *, force_cd: bool) -> Res[None]:
    """Execute the pnpm command with the given arguments.

    Args:
        pnpm_args: The arguments to pass to pnpm.
        force_cd: If True, change the working directory to the mounted volume.

    Returns:
        The exit code of the pnpm command.
    """
    workdir = str(Path.cwd())
    workspace_res = _get_workspaces(workdir)
    if isinstance(workspace_res, Bad):
        return Bad(workspace_res.value)

    workspace, pnpm_workspace = workspace_res.value
    pnpm_command = pnpm_args[0] if pnpm_args else None
    is_cd_command = pnpm_command in PNPM_MOUNTED_COMMANDS
    if not force_cd and (not pnpm_command or not is_cd_command):
        pnpm_res = sub.exec_pnpm(pnpm_args)
        if isinstance(pnpm_res, Bad):
            return pnpm_res
        return Good(None)

    if not Path(f'{workdir}/package.json').exists():
        return issue('missing_package_json', context={
            'workdir': workdir,
            'warning': 'run pnpm commands in directories with package.json',
        })
    pnpm_dir = workdir.replace(workspace, pnpm_workspace, 1)
    return _pnpm(workdir, pnpm_dir, pnpm_args)



from functools import partial
from typing import Any, Callable, TypeVar, cast

from pydantic import ValidationError

from .fp import Bad, Good, OneOf, StopBadIteration
from .issue import Issue

G = TypeVar('G')  # pylint: disable=invalid-name


def issue(  # noqa: WPS211
    message: str,
    description: str | None = None,
    cause: Exception | None = None,
    context: object | None = None,
    include_traceback: bool = True,
) -> OneOf[Issue, Any]:
    """Shortcut to create a Bad OneOf containing an Issue.

    Args:
        message: The issue message.
        description: Optional description.
        cause: Optional exception that caused the issue.
        context: Optional dictionary to provide extra information.
        include_traceback: Defaults to true to provide the stack trace.

    Returns:
        An instance of an `Issue`.
    """
    inst = Issue(message, description, cause, context, include_traceback)
    return Bad(inst)


def one_of(comp: Callable[[], list[G]]) -> OneOf[Any, G]:
    """Handle the "Good" value of a `OneOf`.

    To be used so that we may iterate over OneOf objects that may raise
    the `StopBadIteration` exception.

    Args:
        comp: A lambda function returning an array with a single value.

    Returns:
        A `OneOf` with the value returned from `comp`.
    """
    res = None
    try:
        res = comp()
    except StopBadIteration as ex:
        return cast(Bad, ex.bad)
    except ValidationError as ex:
        return issue('pydantic validation error', cause=ex)
    except Exception as ex:
        return issue('one_of caught exception', cause=ex)
    if res:
        return Good(res[0])
    # LOOK AT ME - you may be here because a mock is not returning a OneOf.
    return issue('one_of empty response - iteration missing a OneOf')


def to_one_of(
    callback: Callable[[], Any],
    message: str,
    context: object | None = None,
) -> OneOf[Issue, int]:
    """Wrap a python call in a `OneOf`.

    Args:
        callback: A lambda function with a simple python statement.
        message: An error message in case the statement raises an exception.
        context: Additional error context.

    Returns:
        A `OneOf` containing an `Issue` if the callback raises an error.
    """
    try:
        callback()
    except Exception as ex:
        return issue(message, cause=ex, context=context)
    return Good(0)


def non_issue(inst: OneOf[Issue, G]) -> G:
    """Obtain the value of the `OneOf` as if it was a Good value.

    Warning: This should only be used provided that we know for sure
    that we are not dealing with a `Bad` value.

    Args:
        inst: A OneOf.

    Returns:
        The value stored in the OneOf.
    """
    # The assert statement can go away with the -O flag.
    assert not inst.is_bad  # noqa: S101
    return cast(G, inst.value)


def _hone(
    msg: str,
    context: object,
    description: str,
    include_traceback: bool,
    iss: Exception,
) -> OneOf[Issue, Any]:
    return issue(
        msg,
        cause=iss,
        context=context,
        description=description,
        include_traceback=include_traceback,
    )


def hone(
    msg: str,
    context: object | None = None,
    description: str | None = None,
    include_traceback: bool = True,
) -> Callable[[Issue], OneOf[Issue, Any]]:
    """Create a function to repackage the issue with a new message and context.

    Args:
        msg: The new message.
        context: The new context.
        description: The new description.
        include_traceback: Whether to include the traceback.

    Returns:
        A function that takes an issue and returns a new issue.
    """
    return partial(_hone, msg, context, description, include_traceback)



import sys
from pathlib import Path

from .one_of import Good, Issue, OneOf, issue


def read_file(filename: str | None) -> OneOf[Issue, str]:
    """FP version of open to read the contents of a file.

    If `None` is provided it will attempt to read from `sys.stdin`.

    Args:
        filename: The file path to read.

    Returns:
        A `Good` containing the contents of the file.
    """
    if filename is None:
        return Good(sys.stdin.read())
    try:
        with Path.open(Path(filename), encoding='UTF-8') as fp:
            return Good(fp.read())
    except Exception as ex:
        return issue(
            'failed to read file',
            context={'filename': filename},
            cause=ex,
        )


def write_file(filename: str, text: str) -> OneOf[Issue, int]:
    """FP version of open to write to a file.

    Args:
        filename: The file path where contents will be written.
        text: The contents of the file.

    Returns:
        A `Good` containing 0 if the file was written.
    """
    try:
        with Path.open(Path(filename), 'w', encoding='UTF-8') as fp:
            fp.write(text)
    except Exception as ex:
        return issue(
            'failed to write file',
            context={'filename': filename},
            cause=ex,
        )
    return Good(0)


def assert_file_exists(path: str) -> OneOf[Issue, Path]:
    """Assert that a file exists.

    Args:
        path: The path to the file.

    Returns:
        A `Good` containing the `Path` if the file exists.
    """
    path_inst = Path(path)
    if path_inst.exists():
        return Good(path_inst)
    return issue('file does not exist', context={'path': path})



import logging

from m.core import Issue, OneOf, one_of
from m.log.misc import default_record_fmt, format_context, format_location

from m import git

from ..types import EnvVars, Message, ProviderModule


def env_vars() -> OneOf[Issue, EnvVars]:
    """Obtain basic environment variables in a local environment.

    Returns:
        An `EnvVars` instance if successful.
    """
    return one_of(
        lambda: [
            EnvVars(
                git_branch=git_branch,
                git_sha=git_sha,
            )
            for git_branch in git.get_branch()
            for git_sha in git.get_current_commit_sha()
        ],
    )


def log_format(
    formatter: logging.Formatter,
    record: logging.LogRecord,
    show_traceback: bool,
    debug_python: bool,
) -> str:
    """Format a log record using the functions provided in this module.

    This function is meant to be used by a log formatter. See more info


    Args:
        formatter: A log formatter instance.
        record: A log record.
        show_traceback: If true, display the python stack trace.
        debug_python: If true, display the location of the record's origin.

    Returns:
        A formatted string.
    """
    record_dict = record.__dict__

    open_b = record_dict.get('open_block')
    if open_b:
        name, desc = open_b
        return f'>>> [{name}]: {desc}'.rstrip()

    if record_dict.get('close_block'):
        return ''

    indent = len(record.levelname) + 3
    context = format_context(record, indent, show_traceback=show_traceback)
    if not record.msg:
        return context[1:]

    ci_info = record_dict.get('ci_info', Message(msg=record.msg))
    msg_info = format_location([ci_info.file, ci_info.line, ci_info.col])
    loc = (
        format_location([record.pathname, f'{record.lineno}'], 'bold_aqua')
        if debug_python
        else ''
    )

    return default_record_fmt(
        record,
        formatter.formatTime(record, formatter.datefmt),
        f'{msg_info}{loc}',
        context,
    )


tool = ProviderModule(
    ci=False,
    env_vars=env_vars,
    formatter=log_format,
)



from m.core import Good, Res, issue

YES_NO = ('yes', 'no')


def is_yes(user_response: str) -> bool:
    """Assert the user response is "yes".

    Args:
        user_response: user's response

    Returns:
        True if the string is "yes".
    """
    return user_response == 'yes'


def assert_branch(branch: str, step: str) -> Res[tuple[str, str]]:
    """Assert that a release step is done in the proper branch.

    This can only happen in `release/x.y.z` or `hotfix/x.y.z`.

    Args:
        branch: branch name to verify.
        step: the release step name.

    Returns:
        An Issue if the current branch is not a release/hotfix else the
        release type and version to release/hotfix.
    """
    valid_prefix = ('release/', 'hotfix/')
    if branch.startswith(valid_prefix):
        parts = branch.split('/')
        return Good((parts[0], parts[1]))
    return issue(
        f'{step}_release can only be done from a release/hotfix branch',
        context={
            'current_branch': branch,
            'expected': 'release/x.y.z or hotfix/x.y.z',
        },
    )



import enum
from dataclasses import dataclass, field
from typing import Callable, Dict, List, Optional

from ....core import Issue, OneOf


class ExitCode(enum.Enum):
    """One of the possible exit codes."""

    ok = 0
    error = 1
    needs_readjustment = 2


@dataclass
class Violation:
    """An error/warning/message provided by a compiler or linter."""

    rule_id: str
    message: str
    line: int
    column: int
    file_path: str


@dataclass
class FileReport:
    """Collection of violations triggered in a single file."""

    file_path: str
    violations: List[Violation]


@dataclass
class RuleInfo:
    """Stats on a rule."""

    rule_id: str
    violations: List[Violation]
    found: int
    allowed: int = 0
    ignored: bool = False


@dataclass
class ProjectStatus:
    """Overview of a project."""

    status: ExitCode
    payload: str
    rules: Dict[str, RuleInfo]
    files: Dict[str, List[Violation]]
    total_found: int = field(init=False)
    total_allowed: int = field(init=False)
    error_msg: str = field(init=False)

    def __post_init__(self):
        """Compute the rest of non initialized variables."""
        rules = self.rules.values()
        total_found = sum((s.found for s in rules if not s.ignored))
        total_allowed = sum((s.allowed for s in rules if not s.ignored))
        error_msg = ''
        if self.status == ExitCode.error:
            diff = sum((
                d
                for s in rules
                for d in (s.found - s.allowed,)
                if d > 0 and not s.ignored
            ))
            error_msg = f'{diff} extra errors were introduced'
        elif self.status == ExitCode.needs_readjustment:
            diff = total_allowed - total_found
            error_msg = f'{diff} errors were removed - lower error allowance'

        self.total_found = total_found  # noqa: WPS601
        self.total_allowed = total_allowed  # noqa: WPS601
        self.error_msg = error_msg  # noqa: WPS601


@dataclass
class Configuration:
    """Options to control the output of the post processor."""

    max_lines: int = 5
    full_message: bool = False
    ignore_error_allowance: bool = False
    file_regex: Optional[str] = None
    file_prefix: Optional[str] = None


Transform = Callable[[str], OneOf[Issue, List[FileReport]]]



from m.cli import BaseModel, command


class Arguments(BaseModel):
    """Print a bash snippet of setup instructions.

    This snippet can be used in a `.bashrc` file to setup the environment.
    It will also provide aliases that call other `m` commands to help
    facilitate the development in a devcontainer workflow.
    """


@command(
    help='print out a bash snippet that exports variables',
    model=Arguments,
)
def run() -> int:
    from m.devcontainer.env import devcontainer_env_vars

    snippet = [
        devcontainer_env_vars().to_bash(),
        'alias pnpm="m devcontainer pnpm"',
        'alias np="m devcontainer pnpm"',
        "alias cd='HOME=$MDC_WORKSPACE cd'",
        'function prompter() { export PS1="$(m devcontainer prompter)"; }',
        'export PROMPT_COMMAND=prompter',
        'export VIRTUAL_ENV="$MDC_VENV_WORKSPACE"',
        'export PATH="$VIRTUAL_ENV/bin:$PATH"',
        'if [ ! -d "$VIRTUAL_ENV" ]; then',
        '  echo "NOTICE: creating virtual environment $VIRTUAL_ENV"',
        '  python3 -m venv "$VIRTUAL_ENV"',
        'fi',
        '. "$VIRTUAL_ENV/bin/activate"',
    ]
    print('\n'.join(snippet))  # noqa: WPS421
    return 0



from dataclasses import dataclass
from typing import Any, Optional

from m.core import Res, http
from pydantic import BaseModel

from .request import request

HttpMethod = http.HttpMethod


def _repos(owner: str, repo: str, *endpoint: str) -> str:
    return '/'.join(['', 'repos', owner, repo, *endpoint])


def create_release(
    token: str,
    owner: str,
    repo: str,
    version: str,
    branch: str | None = None,
) -> Res[Any]:
    """Send a payload to create a release in github.

    Args:
        token: A github PAT.
        owner: The owner of the repo.
        repo: The name of the repo.
        version: The version that marks the release.
        branch: Optional branch to tag, defaults to the default branch.

    Returns:
        The Github response after a release is created.
    """
    endpoint = _repos(owner, repo, 'releases')
    base = 'https://github.com'
    link = f'{base}/{owner}/{repo}/blob/master/CHANGELOG.md#{version}'
    payload = {
        'tag_name': version,
        'name': f'{version}',
        'body': f'**See [CHANGELOG]({link}).**',
        'draft': False,
        'prerelease': False,
    }
    if branch:
        payload['target_commitish'] = branch
    return request(token, endpoint, HttpMethod.post, payload)


class GithubPullRequest(BaseModel):
    """Data needed to create a pull request."""

    title: str
    body: str
    head: str
    base: str


def create_pr(
    token: str,
    owner: str,
    repo: str,
    pr_info: GithubPullRequest,
) -> Res[Any]:
    """Send a payload to create a pull request in github.

    Args:
        token: A github PAT.
        owner: The owner of the repo.
        repo: The name of the repo.
        pr_info: The pull request information.

    Returns:
        The Github response if successful.
    """
    endpoint = _repos(owner, repo, 'pulls')
    return request(token, endpoint, HttpMethod.post, pr_info.model_dump())


def merge_pr(
    token: str,
    owner: str,
    repo: str,
    pr_number: int,
    commit_title: str | None,
) -> Res[Any]:
    """Send a payload to merge a pull request in github.

    Args:
        token: A github PAT.
        owner: The owner of the repo.
        repo: The name of the repo.
        pr_number: The number of the pull request.
        commit_title: An optional commit title to use when merging.

    Returns:
        The payload provided by Github if successful.
    """
    endpoint = _repos(owner, repo, 'pulls', str(pr_number), 'merge')
    payload = {'commit_title': commit_title} if commit_title else {}
    return request(token, endpoint, HttpMethod.put, payload)


@dataclass
class GithubShaStatus:
    """Data needed to create a pull request."""

    sha: str
    context: str
    state: str
    description: str
    url: Optional[str] = None


def commit_status(
    token: str,
    owner: str,
    repo: str,
    sha_info: GithubShaStatus,
) -> Res[Any]:
    """Set a status for a sha.

    The valid states are::

    - pending
    - success
    - failure
    - error

    Args:
        token: A github PAT.
        owner: The owner of the repo.
        repo: The name of the repo.
        sha_info: An instance of a `GithubShaStatus`.

    Returns:
        The response from Github after setting a commit status.
    """
    endpoint = _repos(owner, repo, 'statuses', sha_info.sha)
    payload = {
        'context': sha_info.context,
        'state': sha_info.state,
        'description': sha_info.description,
    }
    if sha_info.url:
        payload['target_url'] = sha_info.url
    return request(token, endpoint, HttpMethod.post, payload)



from m.cli import Arg, BaseModel, command, run_main


class Arguments(BaseModel):
    """Remove empty npm tags.

        $ m npm clean_tags @scope/package

    When packages are removed there will be empty tags that point to nothing.
    This command will find those empty tags and remove them.
    """

    package_name: str = Arg(
        help='name of the npm package',
        positional=True,
        required=True,
    )


@command(
    help='remove empty npm tags',
    model=Arguments,
)
def run(arg: Arguments):
    from m.npm.clean_tags import clean_tags
    return run_main(lambda: clean_tags(arg.package_name))



from m.cli import Meta

meta = Meta(
    help='continuous integration utilities',
    description="""
        The following commands provide local support in the continuous
        integration flow.
    """,
)



from m.cli import (
    Arg,
    BaseModel,
    command,
    create_json_handler,
    create_yaml_handler,
    run_main,
)


class Arguments(BaseModel):
    """Create the [m_dir]/.m/env.list file."""

    pretty: bool = Arg(
        default=False,
        help='format json payload with indentation',
    )
    yaml: bool = Arg(
        default=False,
        help='use yaml format',
    )
    m_dir: str = Arg(
        help='m project directory',
        positional=True,
        required=True,
    )


@command(
    help='create a list of env variables',
    model=Arguments,
)
def run(arg: Arguments) -> int:
    from m.ci.m_env import write_m_env_vars

    result_handler = (
        create_yaml_handler(arg.pretty)
        if arg.yaml
        else create_json_handler(arg.pretty)
    )
    return run_main(
        lambda: write_m_env_vars(arg.m_dir),
        result_handler=result_handler,
    )



from collections import OrderedDict
from enum import Enum
from typing import Any

import yaml

tag_prefix = 'tag:yaml.org,2002'


class _SafeDumper(yaml.SafeDumper):
    """Extends SafeDumper to handle enums."""

    # Argument needs to be called `data` because that's the original signature
    def represent_data(self, data: Any):  # noqa: WPS110
        """Handle special case when dealing with string enums.

        See https://github.com/yaml/pyyaml/issues/51#issuecomment-734782475.

        Args:
            data: A python object.

        Returns:
            A dumper node object.
        """
        if isinstance(data, Enum) and isinstance(data, str):
            return self.represent_data(data.value)
        return super().represent_data(data)


def _ordered_dict_presenter(dumper: yaml.SafeDumper, py_data: OrderedDict):
    return dumper.represent_mapping(f'{tag_prefix}:map', py_data.items())


def _tuple_presenter(dumper: yaml.SafeDumper, py_data: tuple):
    return dumper.represent_sequence(f'{tag_prefix}:seq', py_data)


def _str_presenter(dumper: yaml.SafeDumper, py_data: str):
    if '\n' in py_data:  # check for multiline string
        # tab characters do not play along with yaml
        # found this when trying to print the output of `git status`.
        return dumper.represent_scalar(
            f'{tag_prefix}:str',
            py_data.replace('\r\n', '\n').replace('\t', '    '),
            style='|',
        )
    return dumper.represent_scalar(f'{tag_prefix}:str', py_data)


yaml.add_representer(
    OrderedDict,
    _ordered_dict_presenter,
    Dumper=yaml.SafeDumper,
)
yaml.add_representer(tuple, _tuple_presenter, Dumper=yaml.SafeDumper)
yaml.add_representer(str, _str_presenter, Dumper=yaml.SafeDumper)


def dumps(py_data: Any) -> str:
    """Dump data as yaml using the safe_dump method.

    Args:
        py_data: Any object that may be serialized.

    Returns:
        A yaml serialized string.
    """
    return yaml.dump(py_data, Dumper=_SafeDumper)



import os


def color_disabled() -> bool:
    """Check for environment variable NO_COLOR.

    Utilities that format strings with color should call this function
    to determine if the user wants color in the final output.

    Returns:
        True if it exists and its set to true or 1.
    """
    return os.environ.get('NO_COLOR', 'false') in {'true', '1', 'True'}



from m.log import Logger


def greet(img_name: str, img_version: str, changelog_url: str | None) -> None:
    """Log an "info" message to the console stating that the container is ready.

    Args:
        img_name: The name of the image.
        img_version: The version of the image.
        changelog_url: The URL to the changelog.
    """
    logger = Logger(__name__)
    context = {
        'name': img_name,
        'version': img_version,
        'TIP': 'set `DEBUG`/`DEBUG_M_LOGS` to true to display debugging info',
    }
    if not img_version.startswith('0.0.0') and changelog_url:
        context['changelog'] = f'{changelog_url}#{img_version}'
    logger.info('container_ready', context=context)




import json
from contextlib import suppress
from functools import partial
from typing import Any, Callable

from m.color import highlight_json, highlight_yaml
from m.core import Issue, yaml
from m.log import Logger

logger = Logger('m.cli.cli_helpers')


def _issue_handler(use_warning: bool, iss: Issue):
    """Log an Issue instance.

    Provides the user_warning argument so that CI tools like Teamcity won't
    fail the job.

    Args:
        use_warning: Uses a warning log instead of an error.
        iss: The Issue instance.
    """
    if use_warning:
        logger.warning(iss.message, iss)
    else:
        logger.error(iss.message, iss)


def _json_handler(pretty: bool, inst: Any):
    """Use the print function to display the result as json.

    Args:
        pretty: If true, it highlights and indents the json string.
        inst: A python object.
    """
    if inst is not None:
        msg = inst
        if pretty:
            with suppress(Exception):
                msg = highlight_json(json.dumps(inst, indent=2))
        else:
            with suppress(Exception):
                msg = json.dumps(inst, separators=(',', ':'))
        print(msg)  # noqa: WPS421


def _yaml_handler(pretty: bool, inst: Any):
    """Use the print function to display the result as yaml.

    Args:
        pretty: If true, it highlights the yaml string.
        inst: A python object.
    """
    if inst is not None:
        msg = inst
        if pretty:
            with suppress(Exception):
                msg = highlight_yaml(yaml.dumps(inst))
        else:
            with suppress(Exception):
                msg = yaml.dumps(inst)
        print(msg)  # noqa: WPS421


def create_issue_handler(use_warning: bool) -> Callable[[Issue], None]:
    """Generate a function to log an issue.

    Args:
        use_warning: Uses a warning log instead of an error.

    Returns:
        A function that uses the arguments provided.
    """
    return partial(_issue_handler, use_warning)


def create_json_handler(pretty: bool) -> Callable[[Any], None]:
    """Generate a function to display a value to stdout as json.

    Args:
        pretty: If true, it formats with indentation of 2 spaces.

    Returns:
        A function that uses the arguments provided.
    """
    return partial(_json_handler, pretty)


def create_yaml_handler(pretty: bool) -> Callable[[Any], None]:
    """Generate a function to display a value to stdout as yaml.

    Args:
        pretty: If true, it highlights the output.

    Returns:
        A function that uses the arguments provided.
    """
    return partial(_yaml_handler, pretty)


def create_dict_handler(pretty: bool, as_yaml: bool) -> Callable[[Any], None]:
    """Create a json or yaml handler.

    Args:
        pretty: If true, t highlights the output.
        as_yaml: If true, it formats using yaml.

    Returns:
        A function that uses the arguments
    """
    return (
        create_yaml_handler(pretty)
        if as_yaml
        else create_json_handler(pretty)
    )



from m.cli import Arg, RemainderArgs, command, run_main
from pydantic import BaseModel


class Arguments(BaseModel):
    """Run pnpm in the devcontainer.

    This command is meant to be used as an alias for `pnpm` in a devcontainer.
    If you want view the help for the `pnpm` command you can run::

        command pnpm --help

    Depending on the pnpm command that we want to run, this command will change
    the working directory to the mounted volume and execute the pnpm command.

    Currently the main commands that will be executed in the mounted volume
    are

    - add
    - install
    - remove
    - uninstall

    If you need to run other commands in the mounted volume you can use the
    `--force-cd` flag.
    """

    force_cd: bool = Arg(default=False, help='force `cd` to the mounted volume')

    pnpm_args: list[str] = RemainderArgs(help='arguments to pass to pnpm')


@command(
    help='run pnpm in a devcontainer',
    model=Arguments,
)
def run(arg: Arguments):
    from m.devcontainer.pnpm import run_pnpm

    return run_main(
        lambda: run_pnpm(arg.pnpm_args, force_cd=arg.force_cd),
        result_handler=lambda _: None,
    )



from m.cli import Arg, ArgProxy, BaseModel, command, run_main
from m.core.io import env


class Arguments(BaseModel):
    r"""Create a commit status.

    - https://docs.github.com/en/rest/reference/repos#create-a-commit-status

    example::

        $ m github status \
            --owner jmlopez-rod \
            --repo pysync \
            --sha [sha] \
            --context github-check \
            --state pending \
            --description 'running checks'

    The `owner` option defaults to the value of the environment variable
    `GITHUB_REPOSITORY_OWNER`.
    """

    owner: str = Arg(
        default=env('GITHUB_REPOSITORY_OWNER'),
        help='repo owner',
    )
    repo: str = Arg(
        help='repo name',
        required=True,
    )
    sha: str = Arg(
        help='commit sha',
        required=True,
    )
    context: str = Arg(
        help='unique identifier for the status (a name?)',
        required=True,
    )
    state: str = ArgProxy(
        '--state',
        required=True,
        choices=['error', 'failure', 'pending', 'success'],
        help='the state of the status',
    )
    description: str = Arg(
        help='a short description of the status',
        required=True,
    )
    url: str | None = Arg(
        help='URL to associate with this status',
    )


@command(
    help='create a commit status',
    model=Arguments,
)
def run(arg: Arguments, arg_ns) -> int:
    from m.github.api import GithubShaStatus, commit_status
    return run_main(
        lambda: commit_status(
            arg_ns.token,
            arg.owner,
            arg.repo,
            GithubShaStatus(
                sha=arg.sha,
                context=arg.context,
                state=arg.state,
                description=arg.description,
                url=arg.url,
            ),
        ),
    )



from m.cli import ArgProxy, BaseModel, command, run_main
from m.core.io import env


class Arguments(BaseModel):
    """Start the release process.

    Depending on the workflow the `m` configuration is using we
    will be required to be working on the `master` or `develop` branch.

    It may also require input from the developer to proceed with
    certain operations.
    """

    github_token: str = ArgProxy(
        '--github-token',
        default=env('GITHUB_TOKEN'),
        help='Github PAT (default: env.GITHUB_TOKEN)',
    )


@command(
    help='start the release process',
    model=Arguments,
)
def run(arg: Arguments):
    from m.ci.start_release import start_release

    return run_main(lambda: start_release(arg.github_token, hotfix=False))



from m.cli import Arg, BaseModel, command


class Arguments(BaseModel):
    """Report an error.

    example::

        ~$ m message error 'this is an error'
        ::error::this is an error
        ~$ echo $?
        1

    The procedure exits with non-zero code.
    """

    message: str = Arg(
        help='error message',
        positional=True,
        required=True,
    )
    file: str | None = Arg(  # noqa: WPS110 - required by Github
        aliases=['f', 'file'],
        help='filename where error occurred',
    )
    line: str | None = Arg(
        aliases=['l', 'line'],
        help='line where error occurred',
    )
    col: str | None = Arg(
        aliases=['c', 'col'],
        help='column where error occurred',
    )


@command(
    help='report an error',
    model=Arguments,
)
def run(arg: Arguments):
    from m.log import Logger, Message

    msg = Message(
        msg=arg.message,
        file=arg.file,
        line=arg.line,
        col=arg.col,
    )
    Logger('m.cli').error(msg)
    return 1



from m.cli import Arg, BaseModel, command, run_main


class Arguments(BaseModel):
    """Display a single word representing the current git status.

    example::

        $ m git status
        clean

    statuses::

        unknown
        untracked
        clean
        ahead
        behind
        staged
        dirty
        diverged
        ?

    If you want to check for stashed changes, use the `--check-stashed` flag.
    """

    check_stashed: bool = Arg(
        default=False,
        help='check if there are any stashed changes',
    )


def _get_status(status_desc: tuple[str, str]) -> str:
    return status_desc[0]


@command(
    help='display the current git status',
    model=Arguments,
)
def run(arg: Arguments):
    from m import git

    return run_main(
        lambda: git.get_status(check_stash=arg.check_stashed).map(_get_status),
        print,
    )



from m.cli import Arg, BaseModel, command


class Arguments(BaseModel):
    """Display an npm tag assigned to given version.

    examples::

        ~$ m ci npm_tag 0.0.0-develop.b123
        develop

        ~$ m ci npm_tag 0.0.0-pr1234.b123
        pr1234

        ~$ m ci npm_tag 2.0.1-rc1234.b123
        next

        ~$ m ci npm_tag 2.0.1
        latest
    """

    npm_tag: str = Arg(
        help='npm package version',
        positional=True,
        required=True,
    )


@command(
    help='display an npm tag',
    model=Arguments,
)
def run(arg: Arguments):
    from m.npm.tag import npm_tags

    print(npm_tags(arg.npm_tag)[0])  # noqa: WPS421
    return 0



from m.cli import Arg, command
from m.core import Good
from m.log import Logger
from pydantic import BaseModel


class Arguments(BaseModel):
    """Verify that the required environment variables are set.

    Exists with non-zero exit code if any of the required environment variables
    are not set.
    """

    env_vars: list[str] = Arg(
        help='environment variables to check',
        nargs='+',
        positional=True,
    )


@command(
    help='require env vars',
    model=Arguments,
)
def run(arg: Arguments):
    from m.devcontainer.env import require_env_vars

    env_res = require_env_vars(arg.env_vars)
    if isinstance(env_res, Good):
        return 0

    logger = Logger(__name__)
    logger.error('missing_env_vars', context=env_res.value.model_dump())
    return 2



from m.cli import Arg, BaseModel, command, run_main
from m.cli.handlers import create_json_handler, create_yaml_handler
from m.core.io import env


class Arguments(BaseModel):
    """Retrieve a pull request information.

    example::

        $ m github pr --owner microsoft --repo typescript 44710 | m json
        {
            "headRefName": "ReduceExceptions",
            "headRefOid": "d9ae52cf49732a2d45b6cb7f4069205c88af39eb",
            "baseRefName": "main",
            "baseRefOid": "6452cfbad0afcc6d09b75e0a1e32da1d07e0b7ca",
            "title": "Reduce exceptions",
            "body": "...

    Or use the `--pretty` option to avoid piping.
    """

    owner: str = Arg(
        default=env('GITHUB_REPOSITORY_OWNER'),
        help='repo owner',
    )
    repo: str = Arg(
        help='repo name',
        required=True,
    )
    files: int = Arg(
        default=10,
        help='max number of files to retrieve',
    )
    pretty: bool = Arg(
        default=False,
        help='format json payload with indentation',
    )
    yaml: bool = Arg(
        default=False,
        help='use yaml format',
    )
    pr_number: int = Arg(
        help='the pr number',
        positional=True,
        required=True,
    )


@command(
    help='get information on a pull request',
    model=Arguments,
)
def run(arg: Arguments, arg_ns) -> int:
    from m.github.cli import get_pr_info
    result_handler = (
        create_yaml_handler(arg.pretty)
        if arg.yaml
        else create_json_handler(arg.pretty)
    )
    return run_main(
        lambda: get_pr_info(
            arg_ns.token,
            arg.owner,
            arg.repo,
            arg.pr_number,
            arg.files,
        ),
        result_handler=result_handler,
    )



from typing import Any

from m.cli import Arg, command, validate_json_payload
from pydantic import BaseModel


class Arguments(BaseModel):
    """Format a json payload.

    similar to `python -m json.tool` but instead it uses 2 spaces
    for indentation::

        $ echo '{"a":99}' | m json
        {
          "a": 99
        }

    It is worth noting that if you have access to `jq` or `yq` then
    it should be used instead of `m json`::

        $ echo '{"a":99}' | jq
        {
          "a": 99
        }

    - jq: https://stedolan.github.io/jq/manual/
    - yq: https://mikefarah.gitbook.io/yq/
    """

    payload: Any = Arg(
        default='@-',
        help='json data: @- (stdin), @filename (file), string',
        positional=True,
        validator=validate_json_payload,
    )

    sort_keys: bool = Arg(
        default=False,
        help='sort the output of dictionaries alphabetically by key',
    )


@command(
    help='format json data',
    model=Arguments,
)
def run(arg: Arguments) -> int:
    import json

    from m.color import highlight_json

    json_str = json.dumps(
        arg.payload,
        indent=2,
        sort_keys=arg.sort_keys,
    )
    # Could bypass linter by using run_main but worth doing this
    # especially when we should be using `jq` instead.
    print(highlight_json(json_str))  # noqa: WPS421
    return 0



from m.cli import Arg, BaseModel, command, run_main
from m.core.io import env


class Arguments(BaseModel):
    """Retrieve the latest release.

    example::

        $ m github latest_release --owner microsoft --repo typescript
    """

    owner: str = Arg(
        default=env('GITHUB_REPOSITORY_OWNER'),
        help='repo owner',
    )
    repo: str = Arg(
        help='repo name',
        required=True,
    )


@command(
    help='get the latest release for a repo',
    model=Arguments,
)
def run(arg: Arguments, arg_ns) -> int:
    from m.github.cli import get_latest_release
    return run_main(
        lambda: get_latest_release(
            arg_ns.token,
            arg.owner,
            arg.repo,
        ), print,
    )



from re import sub
from typing import Any, Callable, TypeVar

from pydantic import BaseModel, ConfigDict, TypeAdapter

from .core import Good, Res, hone, issue, one_of
from .core.rw import assert_file_exists
from .core.yaml_fp import read_yson

GenericModel = TypeVar('GenericModel')


def to_camel(snake_case: str) -> str:
    """Transform a string in snake_case to camel case.

    Args:
        snake_case: string in snake case.

    Returns:
        string in camel case.
    """
    s = sub('(_|-)+', ' ', snake_case).title().replace(' ', '')
    return ''.join([s[0].lower(), s[1:]])


class CamelModel(BaseModel):
    """Allows models to be defined with camel case properties.

    See:
        https://medium.com/analytics-vidhya/camel-case-models-with-fast-api-and-pydantic-5a8acb6c0eee
    """

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
    )


def parse_model(model: type[GenericModel], model_data: Any) -> Res[GenericModel]:
    """Parse a python object using pydantics TypeAdapter.

    Args:
        model: The class to create an instance of.
        model_data: The data to parse.

    Returns:
        A `OneOf` with the model or an issue.
    """
    try:
        return Good(TypeAdapter(model).validate_python(model_data))
    except Exception as ex:
        return issue('parse_model_failure', cause=ex)


DataTransformer = Callable[[Any], Res[Any]]


def load_model(
    model: type[GenericModel],
    file_path: str,
    transform: DataTransformer | None = None,
) -> Res[GenericModel]:
    """Load a model from a json or yaml file.

    Args:
        model: The class to create an instance of.
        file_path: The path to the file.
        transform: A function to transform the data before creating the model.

    Returns:
        A `OneOf` with the model or an issue.
    """
    context = {'file_path': file_path, 'model': str(model)}
    transform_function = transform or Good
    return one_of(
        lambda: [
            model_inst
            for _ in assert_file_exists(file_path)
            for model_data in read_yson(file_path)
            for transformed_data in transform_function(model_data)
            for model_inst in parse_model(model, transformed_data)
        ],
    ).flat_map_bad(hone('pydantic.load_model_failure', context))



from m.core import Good, Res, issue, one_of
from m.log import Logger

from .. import git
from . import config as cfg_mod
from .config import Config

logger = Logger('m.ci.assert_branch')


def _verify_branch(
    config: Config,
    branch: str,
    assertion_type: str,
) -> Res[int]:
    workflow = config.workflow
    logger.info(f'verifying "{branch}" branch', {
        'workflow': str(workflow),
    })

    if config.uses_free_flow():
        return issue(
            'The free-flow workflow does not support releases',
            include_traceback=False,
        )
    req_branch = None
    if config.uses_m_flow():
        req_branch = config.m_flow.master_branch
    elif config.uses_git_flow():
        flow = config.git_flow
        # releases are made from develop branch
        # hotfixes are made from master branch
        req_branch = (
            flow.develop_branch
            if assertion_type == 'release'
            else flow.master_branch
        )
    if not req_branch:
        return issue(
            'invalid m workflow',
            context={'workflow': str(workflow)},
            include_traceback=False,
        )
    if branch != req_branch:
        return issue(
            f"invalid branch for '{assertion_type}' using {config.workflow}",
            context={'branch': branch, 'requiredBranch': req_branch},
            include_traceback=False,
        )
    return Good(0)


def assert_branch(assertion_type: str, m_dir: str) -> Res[Config]:
    """Make sure git is using the correct branch based on the workflow.

    Args:
        assertion_type:
            Either 'release' or 'hotfix'.
        m_dir:
            The directory for the m configuration.

    Returns:
        A OneOf containing `None` or an `Issue`.
    """
    return one_of(lambda: [
        config
        for config in cfg_mod.read_config(m_dir)
        for branch in git.get_branch()
        for _ in _verify_branch(config, branch, assertion_type)
    ])



from m.cli import Arg, BaseModel, command, run_main
from m.core.io import env


class Arguments(BaseModel):
    r"""Create a release in Github.

    https://docs.github.com/en/rest/reference/repos#create-a-release

    example::

        $ m github release \
            --owner jmlopez-rod \
            --repo pysync \
            --version 1.0.0

    The `owner` option defaults to the value of the environment variable
    `GITHUB_REPOSITORY_OWNER`.
    """

    owner: str = Arg(
        default=env('GITHUB_REPOSITORY_OWNER'),
        help='repo owner',
    )
    repo: str = Arg(
        help='repo name',
        required=True,
    )
    version: str = Arg(
        help='version to release',
        required=True,
    )
    branch: str | None = Arg(
        help='The branch where the git tag will be created',
    )


@command(
    help='create a github release',
    model=Arguments,
)
def run(arg: Arguments, arg_ns) -> int:
    from m.github.api import create_release
    return run_main(
        lambda: create_release(
            arg_ns.token,
            arg.owner,
            arg.repo,
            arg.version,
            arg.branch,
        ),
    )



import logging
from functools import partial
from typing import Callable, cast

from m.core import Good, Res
from m.core.issue import Issue

from .ci_tools.types import Message

PROMPT = 35


class PromptLogger(logging.Logger):
    """Adding a prompt level to the logger."""

    def __init__(self, name, level=logging.NOTSET):
        """Override the Logger init function.

        Args:
            name: Loggers name.
            level: An optional level.
        """
        super().__init__(name, level)
        logging.addLevelName(PROMPT, 'PROMPT')

    def prompt(self, msg, *args, **kwargs) -> None:
        """Log 'msg % args' with severity 'PROMPT'.

        Args:
            msg: The message to log
            args: arguments to use in the message replacement.
            kwargs: The keywords arguments to provide extra information.
        """
        if self.isEnabledFor(PROMPT):
            self._log(PROMPT, msg, args, **kwargs)


logging.setLoggerClass(PromptLogger)


def log_func_wrapper(
    func: Callable,
    msg: str | Message,
    context: dict | Issue | None = None,
    exit_code: int = 0,
) -> Res[int]:
    """Call a logger function with a message and log data.

    Args:
        func: A logger function (logger.info)
        msg: A string or Message containing information about the log.
        context: Data describing the log entry.
        exit_code: Used in cli context to force an exit. Default to 0.

    Returns:
        A OneOf containing 0 to make it easier to call in fp for loops.
        May be overwritten when signaling an error.
    """
    if isinstance(msg, str):
        func(msg, extra={'context': context}, stacklevel=2)
    else:
        func(
            msg.msg,
            extra={'context': context, 'ci_info': msg},
            stacklevel=2,
        )
    return Good(exit_code)


class Logger:  # noqa: WPS230 - loggers have many attributes
    """Wrapper for python Logger to enable inserting logs in fp loops.

    This is also so that we may attach context data for a message.

    We can access the actual logger by using `inst`.
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, name: str):
        """Initialize the Logger.

        Args:
            name: The name of a logger.
        """
        self.inst = logging.getLogger(name)
        self.debug = partial(log_func_wrapper, self.inst.debug)
        self.info = partial(log_func_wrapper, self.inst.info)  # noqa: WPS110
        self.warning = partial(log_func_wrapper, self.inst.warning)
        self.exception = partial(log_func_wrapper, self.inst.exception)
        self.critical = partial(log_func_wrapper, self.inst.critical)
        self.log = partial(log_func_wrapper, self.inst.log)
        self.prompt = partial(
            log_func_wrapper,
            cast(PromptLogger, self.inst).prompt,
        )
        self.error = partial(log_func_wrapper, self.inst.error)

    def open_block(
        self,
        name: str,
        description: str,
        stderr: bool = False,
    ) -> Res[int]:
        """Group log lines.

        Signals the formatter that the next log lines should be placed in a
        group.

        Args:
            name: The name of the block to open.
            description: Not supported.
            stderr: Force message to be displayed in stderr.

        Returns:
            A OneOf containing 0 to make it easier to call in fp for loops.
        """
        func = self.inst.error if stderr else self.inst.info
        func(
            'OPEN_BLOCK',
            extra={'open_block': (name, description)},
            stacklevel=2,
        )
        return Good(0)

    def close_block(
        self,
        name: str,
        stderr: bool = False,
    ) -> Res[int]:
        """Close a group log lines.

        Signals the formatter that the current group of lines should end.

        Args:
            name: The name of the block to close.
            stderr: Force message to be displayed in stderr.

        Returns:
            A OneOf containing 0 to make it easier to call in fp for loops.
        """
        func = self.inst.error if stderr else self.inst.info
        func(
            'CLOSE_BLOCK',
            extra={'close_block': name},
            stacklevel=2,
        )
        return Good(0)

    def error_block(
        self,
        msg: str | Message,
        context: dict | Issue,
    ) -> Res[int]:
        """Display an error block.

        Args:
            msg: The error message to display.
            context: The dict/Issue to display in a block.

        Returns:
            A OneOf containing 0 to make it easier to call in fp for loops.
        """
        self.error(msg)
        self.open_block('error', '', stderr=True)
        self.error('', context)
        self.close_block('error', stderr=True)
        return Good(0)

    def waning_block(
        self,
        msg: str | Message,
        context: dict | Issue,
    ) -> Res[int]:
        """Display a warning block.

        Args:
            msg: The warning message to display.
            context: The dict/Issue to display in a block.

        Returns:
            A OneOf containing 0 to make it easier to call in fp for loops.
        """
        self.warning(msg)
        self.open_block('warning', '', stderr=True)
        self.warning('', context)
        self.close_block('warning', stderr=True)
        return Good(0)



import setuptools

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name='jmlopez-m',
    version='0.25.1',
    author='Manuel Lopez',
    author_email='jmlopez.rod@gmail.com',
    description='m',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/jmlopez-rod/m',
    project_urls={
        'Bug Tracker': 'https://github.com/jmlopez-rod/m/issues',
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    package_dir={'': 'src'},
    scripts=[
        'bin/m',
        'bin/startRelease',
        'bin/startHotfix',
        'bin/reviewRelease',
        'bin/endRelease',
    ],
    install_requires=[
        'typing_extensions',
        'pydantic>=2.1.0',
        'pyyaml',
        'packaging',
        'pygments>=2.14.0',
        'argcomplete',
    ],
    package_data={'m': ['py.typed']},
    packages=setuptools.find_packages(where='src'),
    python_requires='>=3.10',
)



from functools import cmp_to_key
from typing import Any, Callable, List, Tuple

from m.color import color

from .types import Configuration, ExitCode, ProjectStatus, RuleInfo


def _compare_rules(rule_a: RuleInfo, rule_b: RuleInfo) -> int:
    diff = rule_a.found - rule_b.found or rule_a.allowed - rule_b.allowed
    if not diff:
        return 1 if rule_a.rule_id < rule_b.rule_id else -1
    return diff


def _compare_rule_items(
    rule_a: Tuple[str, RuleInfo],
    rule_b: Tuple[str, RuleInfo],
) -> int:
    return _compare_rules(rule_a[1], rule_b[1])


def _align(token: Any, alignment: str) -> Callable[[int], str]:
    return str(token).ljust if alignment == 'l' else str(token).rjust


def rule_info_str(
    rule: RuleInfo,
    config: Configuration,
) -> str:
    """Format a single rule and its violations.

    Args:
        rule: The rule to display.
        config: The post processor configuration.

    Returns:
        A string representation of the rule info.
    """
    allowed = 'IGNORED' if rule.ignored else f'allowed {rule.allowed}'
    rule_id = color(f'{{red}}{rule.rule_id}{{end}}')
    total_found = color(f'{{red}}{rule.found}{{end}}')
    buffer = [f'{rule_id} (found {total_found}, {allowed}):']
    violations = (
        rule.violations
        if config.max_lines == -1
        else rule.violations[:config.max_lines]
    )
    for violation in violations:
        file_path = violation.file_path
        msg, *rest = violation.message.splitlines()
        line = violation.line
        column = violation.column
        file_loc = color(f'{{gray}}{file_path}:{line}:{column}{{end}}')
        buffer.append(f'  {file_loc} - {msg}')
        if rest and config.full_message:
            buffer.extend([f'    {x}' for x in rest])
    if -1 < config.max_lines < len(rule.violations):
        remaining = len(rule.violations) - config.max_lines
        buffer.append(f'  ... and {remaining} more')
    buffer.append('')
    return '\n'.join(buffer)


def format_row(tokens: List[Any], widths: List[int], alignment: str) -> str:
    """Format a row to be displayed in a table.

    Args:
        tokens:
            A list of values to be displayed.
        widths:
            A list of integers of same size as tokens dictating the how many
            spaces to take for a token.
        alignment:
            Either 'l' or 'r' so that the tokens may be aligned on the left
            or right.

    Returns:
        A single string for a row of a table.
    """
    return '  '.join([
        token
        for index, align in enumerate(alignment)
        for token in (_align(tokens[index], align)(widths[index]),)
    ])


def project_status_str(
    project: ProjectStatus,
    celt_config: Configuration,
) -> str:
    """Stringify a `ProjectStatus` instance.

    Args:
        project: The `ProjectStatus` instance.
        celt_config: The post processor configuration.

    Returns:
        The string version of the project status.
    """
    keys = project.rules.keys()
    rules = sorted(project.rules.values(), key=cmp_to_key(_compare_rules))

    if project.status == ExitCode.ok:
        buffer = [
            rule_info_str(rule, celt_config)
            for rule in rules
            if rule.ignored
        ]
        if project.total_found > 0:
            buffer.append(
                color(
                    '{gray}project has ',
                    f'{{red}}{project.total_found} errors',
                    '{gray} to clear',
                ),
            )
        else:
            msg = 'no errors found'
            buffer.append(color(f'{{bold_green}}{msg}'))
        return '\n'.join(buffer)

    buffer = [
        rule_info_str(rule, celt_config)
        for rule in rules
        if rule.found > rule.allowed
    ]

    buffer.append(color('{bold}FILES:'))
    by_file = sorted(project.files.items(), key=lambda t: len(t[1]))
    buffer.extend([
        color(f'  {{gray}}{file_name}:{{end}} found {total}')
        for file_name, violations in by_file
        for total in (len(violations), )
    ])
    buffer.append('')

    c1_w = max((len(x) for x in keys))
    c1_w = max([c1_w, len('rules')])
    c2_w = max((len(str(s.found)) for s in rules))
    c2_w = max([c2_w, len('found')])
    c3_w = max((len(str(s.allowed)) for s in rules))
    c3_w = max([c3_w, len('allowed')])
    widths = [c1_w, c2_w, c3_w]

    key_rule = sorted(
        project.rules.items(),
        key=cmp_to_key(_compare_rule_items),
    )
    buffer.append(format_row(['RULES', 'FOUND', 'ALLOWED'], widths, 'lll'))
    buffer.extend([
        (
            color(
                '{gray}',
                format_row([rule_id, rule.found, rule.allowed], widths, 'lrr'),
            )
            if rule.found == rule.allowed
            else format_row([rule_id, rule.found, rule.allowed], widths, 'lrr')
        )
        for rule_id, rule in key_rule
        if not rule.ignored
    ])
    buffer.append('')
    return '\n'.join(buffer)


def project_stats_json(
    name: str,
    project: ProjectStatus,
) -> str:
    """Stringify a `ProjectStatus` instance.

    Displays as json showing the current total of violations for each rule.

    Args:
        name: The name of the compiler/linter.
        project: The `ProjectStatus` instance.

    Returns:
        The string version of the project status.
    """
    cap_name = name.capitalize()
    buffer = [
        '{',
        f'  "allowed{cap_name}Rules": {{',
    ]

    key_rule = sorted(
        [
            x
            for x in project.rules.items()
            if x[1].found > 0 and not x[1].ignored
        ],
        key=cmp_to_key(_compare_rule_items),
    )

    if key_rule:
        buffer.extend([
            f'    "{rule_id}": {rule.found},'
            for rule_id, rule in key_rule[:-1]
        ])
        rule_id, rule = key_rule[-1]
        buffer.append(f'    "{rule_id}": {rule.found}')

    buffer.extend([
        '  }',
        '}',
    ])
    return '\n'.join(buffer)



from typing import Dict

from ..core import one_of, subprocess
from ..core.fp import OneOf
from ..core.issue import Issue
from ..core.json import parse_json


def get_dist_tags(pkg_name: str) -> OneOf[Issue, Dict[str, str]]:
    """Fetch all the npm tags for a package.

    Args:
        pkg_name: The name of the npm package.

    Returns:
        A `OneOf` containing an `Issue` or map specifying the tags to a
        published version.
    """
    cmd = f'npm info {pkg_name} dist-tags --json'
    return one_of(lambda: [
        tag_map
        for payload in subprocess.eval_cmd(cmd)
        for tag_map in parse_json(payload)
    ])


def remove_dist_tag(pkg_name: str, tag: str) -> OneOf[Issue, str]:
    """Remove an npm tag.

    Args:
        pkg_name: The name of the npm package.
        tag: The tag to remove.

    Returns:
        A `OneOf` containing an `Issue` or the output of the npm command.
    """
    return subprocess.eval_cmd(f'npm dist-tag rm {pkg_name} {tag} --json')



from m.core import Good, Res, hone, json, one_of
from m.pydantic import parse_model
from pydantic import BaseModel

from ..core.types import FileReport, Violation


class Location(BaseModel):
    """A location in a file."""

    row: int
    column: int


class RuffViolation(BaseModel):
    """A violation from ruff."""

    code: str
    filename: str
    location: Location
    message: str


Report = dict[str, list[Violation]]


def _index_violations(violations: list[RuffViolation]) -> Res[Report]:
    report: Report = {}
    for v_item in violations:
        violation = Violation(
            rule_id=v_item.code,
            message=v_item.message,
            line=v_item.location.row,
            column=v_item.location.column,
            file_path=v_item.filename,
        )
        if violation.file_path not in report:
            report[violation.file_path] = []
        report[violation.file_path].append(violation)
    return Good(report)


def read_payload(payload: str) -> Res[list[FileReport]]:
    """Transform a pylint payload to a list of `FileReport` instances.

    Args:
        payload: The raw payload from pylint.

    Returns:
        A `OneOf` containing an `Issue` or a list of `FileReport` instances.
    """
    context = {'suggestion': 'run ruff check --format json'}
    return one_of(lambda: [
        [
            FileReport(file_path=name, violations=violations)
            for name, violations in report.items()
        ]
        for json_payload in json.parse_json(payload)
        for violations in parse_model(list[RuffViolation], json_payload)
        for report in _index_violations(violations)
    ]).flat_map_bad(hone('invalid_ruff_output_payload', context))



from m.cli import Arg, BaseModel, command, run_main


class Arguments(BaseModel):
    """Update the config and changelog files."""

    changelog: str = Arg(
        default='CHANGELOG.md',
        help='CHANGELOG filename',
    )
    m_dir: str = Arg(
        help='m project directory',
        positional=True,
        required=True,
    )
    new_ver: str = Arg(
        help='the new version',
        positional=True,
        required=True,
    )


@command(
    help='modify the config and changelog files',
    model=Arguments,
)
def run(arg: Arguments) -> int:
    from m.ci.release_setup import release_setup
    return run_main(
        lambda: release_setup(
            arg.m_dir,
            None,
            arg.new_ver,
            arg.changelog,
        ),
    )
